-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
--  Copyright IBM Corp. 1999,2006 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/pa_processpendingdeletes.sql, tbsm.sql, tbsm_base31e
-- SID: 1.4.3.36
-- Date Modified: 07/09/28 16:56:23
--
-- Module Description
--   Defines the pa_processPendingDeletes stored procedure

USE Object
GO

---------------------------------------------------------------------------------------------------
-- pa_processPendingDeletes
PRINT 'Creating procedure pa_processPendingDeletes'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'pa_processPendingDeletes' AND type = 'P')
  DROP PROCEDURE pa_processPendingDeletes
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
--  Copyright IBM Corp. 1999,2006 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/pa_processpendingdeletes.sql, tbsm.sql, tbsm_base31e
-- SID: 1.4.3.36
-- Date Modified: 07/09/28 16:56:23
-- Procedure called by propagation agent to handle the cleanup of a set of objects that have been
-- deleted. Upon entry, if the PendingDeleteState table has a nonzero value in the state column,
-- this indicates that the previous call to this procedure did not complete successfully, and the
-- PendingDeleteInProgress table contains the entries that were being worked on. The state value is
-- used to determine which step of the procedure to restart at. As each step is performed in a
-- separate transaction the PendingDeleteState table is updated. At the end of the procedure 
-- PendingDeleteState.state is set to 0 and the PendingDeleteInProgress table is truncated.
CREATE PROCEDURE pa_processPendingDeletes
  @cid        ClassID = NULL,       -- ignored - parm used in v2.1.1
  @id         ObjID = NULL,         -- ignored - parm used in v2.1.1
  @timestamp  VARBINARY(8) = NULL,  -- ignored - parm used in v2.1.1
  @verbose    TINYINT = 0,
  @maxrows    INT = 0               -- max num entries to process
AS
  SET NOCOUNT ON

  DECLARE @now NVARCHAR(40)
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Starting', 0, 1, @now) WITH NOWAIT
  END

  -- Look up logging mode for this procedure from SystemConfiguration table.
  -- If 1 then each procedure call is logged in PendingDeleteLog
  DECLARE @log_mode TINYINT 
  SET @log_mode = 0           
  SELECT @log_mode = CAST(propertyvalue AS TINYINT)
  FROM SystemConfiguration WITH (NOLOCK)
  WHERE groupname    = 'pa_processPendingDeletes'
    AND propertyname = 'log_mode'

  IF @log_mode = 1
  BEGIN
    DECLARE @starttime DATETIME
    SET @starttime = CURRENT_TIMESTAMP
    INSERT INTO PendingDeleteLog ( starttime ) VALUES ( @starttime )
  END

  DECLARE @err INT, @r INT, @rc INT, @obj_cid ClassID, @obj_id ObjID, @sql NVARCHAR(2000), @table NVARCHAR(255),
          @flag_lightweight TINYINT, @task_id ObjID, @state SMALLINT, @delete_rc INT, @ntasks INT

  -- Lookup state of the previous run of this procedure from the PendingDeleteState table.
  -- Create the record if needed.
  SELECT @state = state FROM PendingDeleteState
  IF @state IS NULL
  BEGIN
    SET @state = 0
    INSERT INTO PendingDeleteState ( mtime, state ) VALUES (CURRENT_TIMESTAMP, 0 )
  END
  
  IF @state = 0 -- No previously unfinished PendingDeletes
  BEGIN
    -- Fill up the PendingDeleteInProgress worktable
    SET ROWCOUNT @maxrows
    INSERT INTO PendingDeleteInProgress (timestamp, cid, id, generation, cno, flag_propagation, in_lob)
      SELECT timestamp, cid, id, generation, cno, flag_propagation, in_lob
        FROM PendingDelete
        ORDER BY timestamp,    -- 1nd order is timestamp i.e. order in which the objects where requested for delete
                 generation    -- 2nd order is hiearchy - top level first so user will not be able to locate pending lower ones that haven't been deleted yet
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    SET ROWCOUNT 0
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Could not insert into PendingDeleteInProgress. Error=%d', 3, 1, @now, @err) WITH NOWAIT
      IF @log_mode = 1
        UPDATE PendingDeleteLog SET err = @err, rc = -30000 WHERE starttime = @starttime
      RETURN -30000
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: %d row(s) inserted into PendingDeleteInProgress.', 0, 1, @now, @rc) WITH NOWAIT
      IF @verbose >= 2
        SELECT TOP 1000 * FROM PendingDeleteInProgress
    END
  END

  IF @state = 0 AND @rc = 0 -- no previously unfinished work and no new work
  BEGIN
    IF @log_mode = 1
      UPDATE PendingDeleteLog SET nrows = 0, err = 0, rc = 0, endtime = CURRENT_TIMESTAMP WHERE starttime = @starttime
    GOTO EndOfProc
  END
  ELSE
  BEGIN
    IF @log_mode = 1
    BEGIN
      IF @state >= 0
        SELECT @rc = COUNT(*) FROM PendingDeleteInProgress 
      UPDATE PendingDeleteLog SET nrows = @rc WHERE starttime = @starttime
    END
  END

  -- Standard transaction logic
  DECLARE @tcount INT
  SET @tcount = @@TRANCOUNT

  IF @state > 0 
  BEGIN
    IF @state = 10
      GOTO DeleteStep020
    ELSE IF @state =  20 
      GOTO DeleteStep030
    ELSE IF @state =  30
      GOTO DeleteStep040
    ELSE IF @state =  40
      GOTO DeleteStep050
    ELSE IF @state =  50
      GOTO DeleteStep060
    ELSE IF @state =  60
      GOTO DeleteStep070
    ELSE IF @state =  70
      GOTO DeleteStep080
    ELSE IF @state =  80
      GOTO DeleteStep090
    ELSE IF @state =  90
      GOTO DeleteStep091
    ELSE IF @state =  91
      GOTO DeleteStep100
    ELSE IF @state = 100
      GOTO DeleteStep110
    ELSE IF @state = 110
      GOTO DeleteStep120
    ELSE IF @state = 120
      GOTO DeleteStep130
    ELSE IF @state = 130
      GOTO DeleteStep140
    ELSE IF @state = 140
      GOTO DeleteStep150
    ELSE IF @state = 150
      GOTO DeleteStep160
    ELSE IF @state = 160
      GOTO DeleteStep170
    ELSE IF @state = 170
      GOTO DeleteStep180
    ELSE IF @state = 180
      GOTO DeleteStep190
    ELSE IF @state = 190
      GOTO DeleteStep200
    ELSE IF @state = 200
      GOTO DeleteStep210
    ELSE IF @state = 210
      GOTO DeleteStep220
    ELSE IF @state = 220
      GOTO DeleteStep230
    ELSE IF @state = 230
      GOTO DeleteStep240
    ELSE IF @state = 240
      GOTO DeleteStep250
    ELSE IF @state = 250
      GOTO DeleteStep260
    ELSE IF @state = 260
      GOTO DeleteStep270
    ELSE IF @state = 270
      GOTO DeleteStep280
    ELSE IF @state = 280
      GOTO DeleteStep290
    ELSE IF @state = 290
      GOTO DeleteStep300
    ELSE IF @state = 300
      GOTO DeleteStep310
  END  

-- STEP 010: Collect the ids of the PropagationSlot entries for the deleted objects
DeleteStep010:
  SET @state = 10
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_010
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_010

  TRUNCATE TABLE PendingDeletePStoDelete
  INSERT INTO PendingDeletePStoDelete
    SELECT ps.id
    FROM PropagationSlot AS ps WITH (NOLOCK)
    JOIN PendingDeleteInProgress AS d ON d.cid = ps.OwnerCID AND d.id = ps.OwnerID
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not lookup PropagationSlots for deleted objects. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_010
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Found %d PropagationSlot record(s) for deleted objects.', 0, 1, @now, @rc) WITH NOWAIT
    IF @verbose >= 2
      SELECT TOP 1000 * FROM PendingDeletePStoDelete
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_010
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_010

-- STEP 020: Delete the scheduled tasks for deleted resources
DeleteStep020:
  SET @state = 20
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_020
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_020

  DECLARE EntMgrTask_c CURSOR FOR
    SELECT st._ParentCID, st._ParentID, st.id 
    FROM ScheduledTask_V AS st WITH (NOLOCK)
      JOIN PendingDeleteInProgress AS d ON d.cid = st._ParentCID AND d.id = st._ParentID
  OPEN EntMgrTask_c
  SET @ntasks = 0
  FETCH NEXT FROM EntMgrTask_c INTO @obj_cid, @obj_id, @task_id
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    EXEC delete_STSK @id = @task_id
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Deleted ScheduledTask %d for %s-%d', 0, 1, @now, @task_id, @obj_cid, @obj_id) WITH NOWAIT
    END
    SET @ntasks = @ntasks + 1
    FETCH NEXT FROM EntMgrTask_c INTO @obj_cid, @obj_id, @task_id
  END
  DEALLOCATE EntMgrTask_c
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d tasks(s)', 0, 1, @now, @ntasks) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_020
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_020

-- STEP 030: Mark as deleted the MaintTrueChildEvent records for the deleted resources
DeleteStep030:
  SET @state = 30
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_030
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_030

  -- Mark any MaintTrueChildEvent records as deleted, else they may propagate later
  UPDATE MaintTrueChildEvent SET deleted = 1
  FROM MaintTrueChildEvent AS m
    JOIN PendingDeleteInProgress AS d ON d.cid = m._ChildCID AND d.id = m._ChildID
  WHERE m.deleted = 0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not update MaintTrueChildEvent. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_030
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Marked %d  record(s) in MaintTrueChildEvent as deleted.', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- OA22288 SR: adding cleanup of the MaintTrueState table for resources being deleted
  DELETE FROM mts
  FROM MaintTrueState AS mts
    JOIN PendingDeleteInProgress AS d ON d.cid = mts.cid AND d.id = mts.id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not delete from MaintTrueState. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_030
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d record(s) from MaintTrueState.', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_030
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_030

-- STEP 040: Delete the StagedCHEV records where the deleted resources are parents or children
DeleteStep040:
  SET @state = 40
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_040
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_040

  -- Delete any StagedCHEV Parent records for the deleted objects
  DELETE FROM s FROM StagedCHEV AS s
    JOIN PendingDeleteInProgress AS d ON d.cid = s.ParentCID AND d.id = s.ParentID
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from StagedCHEV parent. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_040
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d record(s) from StagedCHEV Parent', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Delete any StagedCHEV Child records for the deleted objects
  DELETE FROM s FROM StagedCHEV AS s
    JOIN PendingDeleteInProgress AS d ON d.cid = s.ChildCID  AND d.id = s.ChildID
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from StagedCHEV child. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_040
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d record(s) from StagedCHEV Child', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_040
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_040

-- STEP 050: Delete StagedEXCP records for the deleted resources
DeleteStep050:
  SET @state = 50
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_050
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_050

  DELETE FROM s FROM StagedEXCP AS s
    JOIN PendingDeleteInProgress AS d ON d.cid = s.parent_cid AND d.id = s.parent_id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from StagedEXCP. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_050
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d record(s) from StagedEXCP', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_050
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_050

-- STEP 060: Delete StagedMESG records for the deleted resources
DeleteStep060:
  SET @state = 60
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_060
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_060

  DELETE FROM s FROM StagedMESG AS s
    JOIN PendingDeleteInProgress AS d ON d.cid = s.parent_cid AND d.id = s.parent_id
  WHERE COALESCE(s.AlertState, 
                 (SELECT AlertStateID FROM MessageDescription_T d2 WHERE d2.obj_cid = s.parent_cid AND d2.obj_id  = s.parent_id AND d2.MsgID = s.MESSAGE),
                 (SELECT AlertStateID FROM MessageDescription_T d3 WHERE d3.obj_cid = s.parent_cid AND d3.obj_id  = 0 AND d3.MsgID = s.MESSAGE),
                 0) <> 0 -- do not delete informational messages
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from StagedMESG. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_060
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d record(s) from StagedMESG', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_060
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_060

-- STEP 070: Any LOBs linked to the deleted resources needs to have their CurrentMessageAlertStatePri records reset
DeleteStep070:
  SET @state = 70
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_070
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_070

  -- Update CurrentMessageAlertStatePri for all linked LOBs
  UPDATE c SET AlertState = 1, Priority = 2, mesg_id = 0
  FROM PendingDeleteInProgress AS d
    JOIN lob_link AS l ON l.phy_cid = d.cid AND l.phy_id = d.id AND l.src_id = 0
    JOIN CurrentMessageAlertStatePri AS c ON c.cid = 'LOB' AND c.id = l.dst_id
  WHERE c.AlertState > 1
    AND d.in_lob = 1
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not update CurrentMessageAlertStatePri. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_070
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Updated %d LOB(s) in CurrentMessageAlertStatePri.', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_070
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_070

-- STEP 080: All physical resources that have LOBs linked to them get an OBJECT_DELETED message
-- created on them so that someone looking at the LOB will see that the resource is no longer there
DeleteStep080:
  SET @state = 80
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_080
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_080

  -- Create Message_C records for physical resources linked to LOBs
  INSERT INTO Message_C (name, _TIME, _DISCOVERY_IND, _MESSAGE, _REASON_CD, _StateID, _AlertState, _Priority, _ParentCID, _ParentID)
  SELECT
    'OBJECT_DELETED',
    CURRENT_TIMESTAMP,
    'MON',
    'OBJECT_DELETED',
    'RESOURCE ' + d.cid + ':' + CAST(d.id AS NVARCHAR(10)) + ' DELETED AT ' + CONVERT(NVARCHAR(100), CURRENT_TIMESTAMP, 111) + ' ' + CONVERT(NVARCHAR(100), CURRENT_TIMESTAMP, 108),
    0,
    1,
    2,
    d.cid,
    d.id
  FROM PendingDeleteInProgress AS d
  WHERE d.in_lob = 1 
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not insert into Message_C. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_080
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Inserted %d record(s) into Message_C.', 0, 1, @now, @rc) WITH NOWAIT
  END
  
  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_080
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_080

-- STEP 090: Delete exceptions on the deleted resources
DeleteStep090:
  SET @state = 90
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_090
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_090

  -- Delete the Exceptions
  -- This will clear them for LOBs as well
  UPDATE e SET deleted = 1
  FROM Exception_V AS e
    JOIN PendingDeleteInProgress AS d ON d.cid = e._ParentCID AND d.id = e._ParentID
  WHERE e.deleted = 0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not update Exception_V. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_090
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Marked %d Exception_V record(s) as deleted for objects not in LOBs', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_090
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_090

-- STEP 091: Delete messages on the deleted resources (but not the OBJECT_DELETED messages, just created)
DeleteStep091:
  SET @state = 91
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_091
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_091

  -- Delete the Messages
  UPDATE m SET deleted = 1
  FROM Message_V AS m
    JOIN PendingDeleteInProgress AS d ON d.cid = m._ParentCID AND d.id = m._ParentID
  WHERE m.name <> 'OBJECT_DELETED' 
  AND m.deleted = 0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not update Message_V. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_091
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Marked %d Message_V record(s) as deleted for objects.', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_091
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_091


-- STEP 100: Close any open ownership notes for which the delete resources are the only dependents
DeleteStep100:
  SET @state = 100
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_100
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_100

  -- Handle closing any notes that are open due exclusively to events on the deleted objects
  TRUNCATE TABLE PendingDeleteNoteID
  -- Get a list of note ids that are affected by the deletes.  This is any open note that has an entry in the
  -- NoteDependent table for a deleted resource.
  INSERT INTO PendingDeleteNoteID (note_id)
  SELECT DISTINCT nd.noteid
  FROM NoteDependent AS nd
    JOIN Note_V AS n ON n.id = nd.noteid
    JOIN PendingDeleteInProgress AS d ON d.cid = nd.depcid AND d.id = nd.depid 
  WHERE n._Status = 'Open'
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not lookup open notes. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_100
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Found %d open note(s)', 0, 1, @now, @rc) WITH NOWAIT
  END

  IF @rc > 0 -- close open notes
  BEGIN
    -- Find notes that depend on other non-deleted resources
    DELETE i
    FROM PendingDeleteNoteID AS i
      JOIN OwnedEvent AS oe ON oe.noteid = i.note_id
      LEFT OUTER JOIN PendingDeleteInProgress AS d ON d.cid = oe.parentcid AND d.id = oe.parentid
    WHERE d.cid IS NULL
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Could not lookup open notes related to non-deleted resources. Error=%d', 3, 1, @now, @err) WITH NOWAIT
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_100
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Found %d open note(s) related to non-deleted resources', 0, 1, @now, @rc) WITH NOWAIT
    END

    IF EXISTS (SELECT 1 FROM PendingDeleteNoteID)
    BEGIN -- Close remaining notes 
      UPDATE n SET _Status = 'Closed', _Resolution = 'Object Deleted'
      FROM Note_A AS n
        JOIN PendingDeleteNoteID AS i ON i.note_id = n.id
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @err <> 0
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: pa_processPendingDeletes: Could not update note status. Error=%d', 3, 1, @now, @err) WITH NOWAIT
        SET @rc = -(30000 + @state)
        ROLLBACK TRANSACTION pa_processPendingDeletes_100
        GOTO ProcFailed
      END
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: pa_processPendingDeletes: Closed %d open note(s)', 0, 1, @now, @rc) WITH NOWAIT
      END
    END -- Close remaining notes 
  END -- close open notes

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_100
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_100

-- STEP 110: Cleanup the NoteDependent table for the deleted resources by setting status = 3
DeleteStep110:
  SET @state = 110
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_110
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_110

  UPDATE nd SET status = 3
  FROM NoteDependent AS nd
    JOIN PendingDeleteInProgress AS d ON d.cid = nd.depcid AND d.id = nd.depid
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not update NoteDependent. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_110
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Updated %d row(s) in NoteDependent', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_110
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_110

-- STEP 120: Cleanup the OwnedEvent table for the delete resources by setting status = 3
DeleteStep120:
  SET @state = 120
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_120
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_120

  UPDATE oe SET status = 3
  FROM OwnedEvent AS oe
    JOIN PendingDeleteInProgress AS d ON d.cid = oe.parentcid AND d.id = oe.parentid
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Could not update OwnedEvent. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_120
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Updated %d row(s) in OwnedEvent', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_120
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_120

-- STEP 130: Perform processing associated with individual class tables.
DeleteStep130:
  SET @state = 130
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_130
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_130

  -- Perform any processing that is associated with the individual class tables. Currently this
  -- encompasses the following cases:
  -- 1. All _C tables for deleted objects get deleted set to 1.
  -- 2. All LOBs linked to deleted objects have their _DeletedCID, _DeletedID, _DeletedTime, and
  --    _DeletedName columns updated.

  -- The cursor declaration now joins to make sure cid is valid. Because, the cid was getting deleted
  -- from obj_class when the class is dropped (ususally done during un-install of PAKs)
  DECLARE ppd_c CURSOR FOR
    SELECT DISTINCT d.cid, oc.cname + '_C', oc.flags & 1 /* flag_lightweight */
    FROM PendingDeleteInProgress AS d
    JOIN Meta..obj_class_table_BASE AS oc WITH (NOLOCK) ON oc.cid = d.cid
  OPEN ppd_c
  FETCH NEXT FROM ppd_c INTO @obj_cid, @table, @flag_lightweight
  WHILE @@FETCH_STATUS <> -1
  BEGIN

    -- If the class is lightweight we don't do anything
    IF @flag_lightweight = 1
    BEGIN
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: pa_processPendingDeletes: Skipping lightweight class %s', 0, 1, @now, @obj_cid) WITH NOWAIT
      END
      GOTO NextClass
    END

    -- 1. All _C tables for deleted objects get deleted set to 1.
    SET @sql =
    'UPDATE v SET deleted=1, mtime=CURRENT_TIMESTAMP FROM ' + @table + ' AS v ' +
    'JOIN PendingDeleteInProgress AS d ON d.cid=''' + @obj_cid + ''' AND d.id=v.id '
    EXEC @r = sp_executesql @sql
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0 OR @r = 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Failed to update deleted flag in %s. Error=%d', 3, 1, @now, @table, @err) WITH NOWAIT
      DEALLOCATE ppd_c
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_130
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Updated deleted flag on %4d record(s) in %s', 0, 1, @now, @rc, @table) WITH NOWAIT
    END

    -- 2. All LOBs linked to deleted objects have their _DeletedCID, _DeletedID, _DeletedTime, and
    --    _DeletedName columns updated.
    SET @sql =
    'UPDATE lob SET  ' +
      '_DeletedCID = l.dst_cid, ' +
      '_DeletedID = l.dst_id, ' +
      '_DeletedTime = CURRENT_TIMESTAMP, ' +
      '_DeletedName = o.name ' +
    'FROM ' +
      'LineOfBusiness_A AS lob ' +
      'JOIN link AS l ON l.src_cid = ''LOB'' AND l.src_id = lob.id ' +
      'JOIN PendingDeleteInProgress AS d ON d.cid = l.dst_cid AND d.id = l.dst_id ' +
      'JOIN ' + @table + ' AS o ON o.id = l.dst_id ' +
    'WHERE ' +
          'l.link_type = ''LOBL'' ' +
      'AND l.dst_cid = @obj_cid ' +
      'AND l.deleted = 0 '
    EXEC @r = sp_executesql @sql, N'@obj_cid ClassID',
      @obj_cid    = @obj_cid
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0 OR @r = 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Failed to update LOBs from %s. Error=%d', 0, 1, @now, @table, @err) WITH NOWAIT
      DEALLOCATE ppd_c
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_130
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Updated %d LOB(s) from %s', 0, 1, @now, @rc, @table) WITH NOWAIT
    END

    -- 3. All LOBs linked to deleted objects should be deleted from CL_IDCache.  (OA10691)
    SET @sql =
    'DELETE idc FROM ' +
      'CL_IDCache AS idc ' +
      'JOIN LineOfBusiness_A AS lob ON lob.id = idc.oid AND idc.cid = ''LOB'' ' +
      'JOIN link AS l ON l.src_cid = ''LOB'' AND l.src_id = lob.id ' +
      'JOIN PendingDeleteInProgress AS d ON d.cid = l.dst_cid AND d.id = l.dst_id ' +
      'JOIN ' + @table + ' AS o ON o.id = l.dst_id ' +
    'WHERE ' +
          'l.link_type = ''LOBL'' ' +
      'AND l.dst_cid = @obj_cid ' +
      'AND l.deleted = 0 '
    EXEC @r = sp_executesql @sql, N'@obj_cid ClassID',
      @obj_cid    = @obj_cid
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0 OR @r = 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Failed to deleted LOBs from CL_IDCache. Error=%d', 0, 1, @now, @err) WITH NOWAIT
      DEALLOCATE ppd_c
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_130
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Deleted %d LOBs linked to deleted objects from CL_IDCache', 0, 1, @now, @rc) WITH NOWAIT
    END

NextClass:
    FETCH NEXT FROM ppd_c INTO @obj_cid, @table, @flag_lightweight
  END
  DEALLOCATE ppd_c

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_130
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_130

-- STEP 140: Send negating child events from the deleted resources to their non-deleted parents
DeleteStep140:
  SET @state = 140
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_140
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_140

  -- For any of the deleted objects we have to send negating child events to any of the parent
  -- objects that are NOT deleted.
  TRUNCATE TABLE PendingDeleteParent
  INSERT INTO PendingDeleteParent
    SELECT DISTINCT c._ParentCID, c._ParentID, 0
    FROM ChildEvent_V AS c
      JOIN PendingDeleteInProgress AS d ON d.cid = c._ChildCID AND d.id = c._ChildID
    WHERE c.deleted = 0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to find parent objects with child events. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_140
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Found %d parent object(s) with child events', 0, 1, @now, @rc) WITH NOWAIT
  END
  -- Look for deleted parents
  IF @rc > 0
  BEGIN
    DECLARE ppdparent_c CURSOR FOR
      SELECT DISTINCT p.cid, c.cname + '_C' FROM PendingDeleteParent AS p
        JOIN obj_class AS c ON c.cid = p.cid
    OPEN ppdparent_c
    FETCH NEXT FROM ppdparent_c INTO @obj_cid, @table
    WHILE @@FETCH_STATUS <> -1
    BEGIN
      SET @sql =
      'UPDATE p SET deleted = 1 ' +
      'FROM PendingDeleteParent AS p ' +
        'JOIN ' + @table + ' AS c ON c.id = p.id ' +
      'WHERE ' +
        'p.cid = @obj_cid ' +
        'AND c.deleted = 1 '
      EXEC @r = sp_executesql @sql, N'@obj_cid ClassID',
        @obj_cid = @obj_cid
      SELECT @err = @@ERROR, @rc = @@ROWCOUNT
      IF @err <> 0 OR @r = 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: pa_processPendingDeletes: Failed to lookup parent deleted flags from %s. Error=%d', 0, 1, @now, @table, @err) WITH NOWAIT
      END
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: pa_processPendingDeletes: Found %d deleted parent object(s) in %s', 0, 1, @now, @rc, @table) WITH NOWAIT
      END

      FETCH NEXT FROM ppdparent_c INTO @obj_cid, @table
    END
    DEALLOCATE ppdparent_c
  END

  IF EXISTS (SELECT * FROM PendingDeleteParent WHERE deleted = 0)
  BEGIN
    -- Compute table of child events to delete
    TRUNCATE TABLE PendingDeleteCHEV
    INSERT INTO PendingDeleteCHEV
    SELECT
      c.id,
      c._ParentCID,
      c._ParentID,
      c._ChildCID,
      c._ChildID,
      c._AlertState,
      c._Priority,
      c._nCHVH
    FROM
      ChildEvent_V AS c
      JOIN PendingDeleteParent AS p ON p.cid = c._ParentCID AND p.id = c._ParentID
      JOIN PendingDeleteInProgress AS d ON d.cid = c._ChildCID AND d.id = c._ChildID
    WHERE
      c.deleted = 0
      AND p.deleted = 0
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Failed to compute child event(s) for deletion. Error=%d', 3, 1, @now, @err) WITH NOWAIT
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_140
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Found %d child event(s) to delete', 0, 1, @now, @rc) WITH NOWAIT
      IF @verbose >= 2
        SELECT 'PendingDeleteCHEV' AS 'PendingDeleteCHEV', * FROM PendingDeleteCHEV
    END
    UPDATE c SET deleted = 1
    FROM ChildEvent_V AS c JOIN PendingDeleteCHEV AS cd ON cd.id = c.id
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Failed to mark ChildEvent_V for deletion. Error=%d', 3, 1, @now, @err) WITH NOWAIT
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_140
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Marked %d ChildEvent_V records(s) as deleted', 0, 1, @now, @rc) WITH NOWAIT
    END

  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_140
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_140

-- STEP 150: Mark as deleted all containment records for the deleted resources
DeleteStep150:
  SET @state = 150
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_150
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_150

  -- Mark all containment record(s) for the deleted objects
  UPDATE c
    SET deleted=1
  FROM
    containment AS c
    JOIN PendingDeleteInProgress AS d ON d.cid = c.dst_cid AND d.id = c.dst_id
  WHERE c.deleted=0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to mark dst containment record(s) for deletion. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_150
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Marked %d dst containment record(s) for deletion', 0, 1, @now, @rc) WITH NOWAIT
  END

  UPDATE c
    SET deleted=1
  FROM
    containment AS c
    JOIN PendingDeleteInProgress AS d ON d.cid = c.src_cid AND d.id = src_id
  WHERE c.deleted=0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to mark src containment record(s) for deletion. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_150
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Marked %d src containment record(s) for deletion', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_150
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_150

-- STEP 160: Mark as deleted all link records where the deleted resources are on the dst side
DeleteStep160:
  SET @state = 160
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_160
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_160

  UPDATE l 
    SET deleted=1
  FROM
    link AS l
    JOIN PendingDeleteInProgress AS d ON d.cid = l.dst_cid AND d.id = l.dst_id
  WHERE l.deleted=0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to mark dst link record(s) for deletion. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_160
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Marked %d dst link record(s) for deletion', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_160
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_160

-- STEP 170: Mark as deleted all link records where the deleted resources are on the src side
DeleteStep170:
  SET @state = 170
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_170
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_170

  UPDATE l
    SET deleted=1
  FROM
    link AS l
    JOIN PendingDeleteInProgress AS d ON d.cid = l.src_cid AND d.id = l.src_id
  WHERE l.deleted=0
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to mark src link record(s) for deletion. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_170
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Marked %d src link record(s) for deletion', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_170
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_170

  -- If a LOB is being deleted then all records in lob_link for this object
  -- (whether on the source or destination side) need to be deleted as well.
  -- These are done with two separate delete statements because the performance
  -- when trying to use an or statement is pretty bad (i.e. - table scan).
  IF EXISTS (SELECT 1 FROM PendingDeleteInProgress WHERE cid = 'LOB')
  BEGIN
-- STEP 180: Delete lob_link.dst_id records
DeleteStep180:
SET @state = 180
    IF @tcount = 0
      BEGIN TRANSACTION pa_processPendingDeletes_180
    ELSE
      SAVE TRANSACTION pa_processPendingDeletes_180
    
    -- Delete lob_link.src_id records
    DELETE FROM l
    FROM lob_link AS l
    JOIN PendingDeleteInProgress AS p ON p.cid = 'LOB' AND p.id = l.src_id
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Failed to delete src rows from lob_link. Error=%d', 3, 1, @now, @err) WITH NOWAIT
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_180
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Deleted %d src lob_link records for deleted LOBC links', 0, 1, @now, @rc) WITH NOWAIT
    END

    EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
    IF @rc < 1
    BEGIN
      ROLLBACK TRANSACTION pa_processPendingDeletes_180
      GOTO ProcFailed
    END

    -- Standard transaction logic
    IF @tcount = 0
      COMMIT TRANSACTION pa_processPendingDeletes_180

-- STEP 190: Delete lob_link.dst_id records
DeleteStep190:
SET @state = 190
    IF @tcount = 0
      BEGIN TRANSACTION pa_processPendingDeletes_190
    ELSE
      SAVE TRANSACTION pa_processPendingDeletes_190

    DELETE l
    FROM lob_link AS l
    JOIN PendingDeleteInProgress AS p ON p.cid = 'LOB' AND p.id = l.dst_id
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Failed to delete dst rows from lob_link. Error=%d', 3, 1, @now, @err) WITH NOWAIT
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_190
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Deleted %d dst lob_link records for deleted LOBC links', 0, 1, @now, @rc) WITH NOWAIT
    END

    EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
    IF @rc < 1
    BEGIN
      ROLLBACK TRANSACTION pa_processPendingDeletes_190
      GOTO ProcFailed
    END

    -- Standard transaction logic
    IF @tcount = 0
      COMMIT TRANSACTION pa_processPendingDeletes_190
  END -- IF EXISTS (SELECT 1 FROM PendingDeleteInProgress WHERE cid = 'LOB')

-- STEP 200: Reset the lob_link phy_cid,phy_id columns where they specify deleted resources
DeleteStep200:
SET @state = 200
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_200
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_200
  
  UPDATE l SET phy_cid = NULL, phy_id = 0
    FROM lob_link AS l
    JOIN PendingDeleteInProgress AS p ON p.cid = l.phy_cid AND p.id = l.phy_id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Error updating phy_cid,phy_id in lob_link. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_200
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Updated %d lob_link row(s) for LOB phy_cid', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_200
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_200

-- STEP 210: Delete the GEM_IDLookup records for the deleted resources
DeleteStep210:
SET @state = 210
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_210
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_210

  -- Begin gem processing
  DELETE g FROM GEM_IDLookup AS g JOIN PendingDeleteInProgress AS d ON d.cid = g.CID AND d.id = g.IID
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from GEM_IDLookup. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_210
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d record(s) from GEM_IDLookup', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_210
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_210

-- STEP 220: Reset GEM_LOBLookup.LOBID for the deleted LOBs
DeleteStep220:
SET @state = 220
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_220
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_220

  -- gem lob processing
  UPDATE GEM_LOBLookup SET LOBID = 0 FROM PendingDeleteInProgress AS d
  WHERE d.cid = 'LOB'
    AND d.id = LOBID
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to clear LOBID in GEM_LOBLookup. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_220
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Cleared LOBID in %d row(s) in GEM_LOBLookup', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_220
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_220

-- STEP 230: Reset GEM_LOBLookup.parentLOBID for the deleted LOBs
DeleteStep230:
SET @state = 230
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_230
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_230

  UPDATE GEM_LOBLookup SET parentLOBID = 0
  FROM PendingDeleteInProgress AS d
  WHERE d.cid = 'LOB'
    AND d.id = parentLOBID
    AND parentLOBKey <> 'LOBC'
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to clear parentLOBID in GEM_LOBLookup. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_230
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Cleared parentLOBID in %d row(s) in GEM_LOBLookup', 0, 1, @now, @rc) WITH NOWAIT
  END
  -- End gem processing

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_230
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_230

-- STEP 240: Delete from ABS if tracking of CL instance id is enabled
DeleteStep240:
SET @state = 240
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_240
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_240

  -- Delete from ABS if tracking of CL instance id is enabled
  DECLARE @abs INT
  SET @abs = 0
  SELECT @abs=ISNULL(propertyvalue, 0) FROM SystemConfiguration WHERE groupname='ABS' AND propertyname='bAttribute_cl_instid'
  IF @abs <> 0
  BEGIN 
    TRUNCATE TABLE PendingDeleteAlobCacheData
    -- put values in table
    INSERT INTO PendingDeleteAlobCacheData (iProcess, iClass, iObject, sNameAttribute, sValueAttributeDeleted, 
                                sValueAttributeInserted,  iAction, iDatatype)
      SELECT 1, d.cid, d.id, 'cl_instid', c.value, NULL, 3, 231 
      FROM PendingDeleteInProgress AS d
        JOIN CL_IDCache AS c ON c.cid=d.cid AND c.oid=d.id
      WHERE c.keyword='CL_InstID'
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @rc > 0
    BEGIN
      -- Notify abs of deleted cl_instid values
      EXEC @err = alobPutAttributeNonModel 'PendingDeleteAlobCacheData', 1
      TRUNCATE TABLE PendingDeleteAlobCacheData
      IF @err <> 0
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: pa_processPendingDeletes: Failed to delete CL related records from ABS cache. Error=%d', 3, 1, @now, @err) WITH NOWAIT
        SET @rc = -(30000 + @state)
        ROLLBACK TRANSACTION pa_processPendingDeletes_240
        GOTO ProcFailed
      END
      IF @verbose >= 1
      BEGIN
        SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
        RAISERROR('%s: pa_processPendingDeletes: Deleted %d CL related record(s) from ABS cache', 0, 1, @now, @rc) WITH NOWAIT
      END
    END
  END
  ELSE IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Did not notify ABS of deleted CL_InstID records since SystemConfiguration.bAttribute_cl_instid = 0', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_240
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_240

-- STEP 250: Removed ids of deleted resources from the CL_IDCache table
DeleteStep250:
SET @state = 250
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_250
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_250

  DELETE g FROM CL_IDCache AS g JOIN PendingDeleteInProgress AS d ON d.cid = g.cid AND d.id = g.oid
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from CL_IDCache. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_250
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d record(s) from CL_IDCache', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_250
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_250

-- STEP 260: Delete from CL_LinkRules
DeleteStep260:
SET @state = 260
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_260
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_260

  DELETE lr FROM CL_LinkRules AS lr
    JOIN PendingDeleteInProgress AS d ON (d.cid=lr.src_cid AND d.id=lr.src_id)
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from CL_LinkRules. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_260
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d record(s) from CL_LinkRules', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_260
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_260

-- STEP 270: Decrement the child count in the parent cname tables
DeleteStep270:
SET @state = 270
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_270
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_270

  TRUNCATE TABLE PendingDeleteParentInfo
  INSERT INTO PendingDeleteParentInfo
    SELECT DISTINCT l.src_cid, l.src_id, COUNT(*)
    FROM link AS l
    JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = l.link_type
    JOIN PendingDeleteInProgress AS d ON d.cid = l.dst_cid AND d.id = l.dst_id
    LEFT OUTER JOIN PendingDeleteInProgress AS dp ON dp.cid = l.src_cid AND dp.id = l.src_id
    WHERE dp.cid IS NULL
      AND lt.flag_containment = 1
    GROUP BY l.src_cid, l.src_id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @r = 1 OR @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to compute nchild updates. Error=%d', 3, 1, @now, @table, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_270
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Computed %d nchild update(s)', 0, 1, @now, @rc) WITH NOWAIT
  END

  DECLARE ncld_c INSENSITIVE CURSOR FOR
    SELECT DISTINCT oc.cname + '_C', oc.cid
    FROM PendingDeleteParentInfo AS p
    JOIN obj_class AS oc ON (oc.cid = p.src_cid)
  OPEN ncld_c
  FETCH NEXT FROM ncld_c INTO @table, @obj_cid
  WHILE @@FETCH_STATUS <> -1
  BEGIN
    SET @sql = 
        'UPDATE c SET nchild = CASE WHEN (c.nchild - p.nchild) >= 0 THEN (c.nchild - p.nchild) ELSE 0 END ' +
          'FROM ' + @table + ' AS c ' +  
          'JOIN PendingDeleteParentInfo AS p ON c.id = p.src_id ' + 
          'WHERE p.src_cid = ''' + @obj_cid + ''' AND c.nchild > 0 '

    EXEC @r = sp_executesql @sql
    SELECT @err = @@ERROR, @rc = @@ROWCOUNT
    IF @r = 1 OR @err <> 0
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Failed to update nchild flag in %s. Error=%d', 3, 1, @now, @table, @err) WITH NOWAIT
      DEALLOCATE ncld_c
      SET @rc = -(30000 + @state)
      ROLLBACK TRANSACTION pa_processPendingDeletes_270
      GOTO ProcFailed
    END
    IF @verbose >= 1
    BEGIN
      SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
      RAISERROR('%s: pa_processPendingDeletes: Updated nchild on %4d record(s) in %s', 0, 1, @now, @rc, @table) WITH NOWAIT
    END
    
    FETCH NEXT FROM ncld_c INTO @table, @obj_cid
  END
  DEALLOCATE ncld_c

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_270
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_270

-- STEP 280: Send notifications
DeleteStep280:
SET @state = 280
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_280
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_280

  -- Send notifications of all propagation objects that have just been deleted
  INSERT INTO PendingNotification(type, fields, detail)
  SELECT
    type = et.type,
    fields=
      CONVERT(NVARCHAR, ISNULL(d.cno, 0)) + '|' +
      CONVERT(NVARCHAR, ISNULL(d.id, 0))  + '|' +
      CASE generation WHEN 0 THEN '1' ELSE '0' END,
    ''
  FROM
    PendingDeleteInProgress AS d,
    EventType AS et WITH (NOLOCK)
  WHERE d.flag_propagation = 1
    AND et.name = 'ObjectDeleted'
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to insert into PendingNotification. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_280
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Inserted %d row(s) into PendingNotification', 0, 1, @now, @rc) WITH NOWAIT
  END

  -- Send notifications to console that LOB _DeletedID has been updated (otherwise the console will not refresh to show the _DeletedCID icon)
  -- Note:  For some reason telling the console that DeleteCID was updated doesn't cause a Refresh, but DeleteID does.
  DECLARE @type INT, @cno INT, @aid INT
  SELECT @type = type FROM EventType     WHERE name   = 'ObjAttrUpdated'
  SELECT @cno  = cno  FROM obj_class     WHERE cid    = 'LOB'
  SELECT @aid  = id   FROM id_string_map WHERE string = 'DeletedID'
  INSERT INTO PendingNotification(type, fields, detail)
    SELECT @type, 
           CAST(ISNULL(@cno,0) AS NVARCHAR)+'|'+CAST(lobl.src_id AS NVARCHAR)+'|'+CAST(ISNULL(@aid,0) AS NVARCHAR),
           CAST(d.id AS NVARCHAR)
      FROM PendingDeleteInProgress   AS d
      JOIN link             AS lobl ON lobl.dst_cid = d.cid AND lobl.dst_id = d.id
      WHERE lobl.link_type = 'LOBL' AND lobl.src_cid = 'LOB'
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to insert into PendingNotification for _DeletedID. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_280
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Inserted %d row(s) into PendingNotification for _DeletedID', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_280
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_280

-- STEP 290: Remove PropagationSlots for deleted objects
DeleteStep290:
SET @state = 290
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_290
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_290

  DELETE FROM ps
    FROM PropagationSlot AS ps
    JOIN PendingDeletePStoDelete AS d ON d.id = ps.id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT   
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from PropagationSlot. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_290
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d row(s) from PropagationSlot', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_290
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_290

-- STEP 300: Remove PendingDelete records we've just processed
DeleteStep300:
SET @state = 300
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_300
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_300

  DELETE pd
    FROM PendingDelete  AS pd
    JOIN PendingDeleteInProgress AS t  ON t.cid = pd.cid AND t.id = pd.id
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from PendingDelete. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_300
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d row(s) from PendingDelete', 0, 1, @now, @rc) WITH NOWAIT
  END

  EXEC @rc = pa_logPendingDeleteState @state = @state, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_300
    GOTO ProcFailed
  END

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_300

-- STEP 310: Remove the PendingDeleteInProgress records that we've just processed
DeleteStep310:
SET @state = 310
  IF @tcount = 0
    BEGIN TRANSACTION pa_processPendingDeletes_310
  ELSE
    SAVE TRANSACTION pa_processPendingDeletes_310

  DELETE FROM PendingDeleteInProgress
  SELECT @err = @@ERROR, @rc = @@ROWCOUNT  --- NOTE:  @rc is used on the return to tell the SEL how many were processed.
  IF @err <> 0
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Failed to delete from PendingDeleteInProgress. Error=%d', 3, 1, @now, @err) WITH NOWAIT
    SET @rc = -(30000 + @state)
    ROLLBACK TRANSACTION pa_processPendingDeletes_310
    GOTO ProcFailed
  END
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Deleted %d row(s) from PendingDeleteInProgress', 0, 1, @now, @rc) WITH NOWAIT
  END

  SET @delete_rc = @rc

  EXEC @rc = pa_logPendingDeleteState @state = 0 /* NOT @state since we're finished */, @verbose = @verbose
  IF @rc < 1
  BEGIN
    ROLLBACK TRANSACTION pa_processPendingDeletes_310
    GOTO ProcFailed
  END

  SET @rc = @delete_rc

  -- Standard transaction logic
  IF @tcount = 0
    COMMIT TRANSACTION pa_processPendingDeletes_310

  IF @log_mode = 1
    UPDATE PendingDeleteLog SET err = 0, rc = @rc, endtime = CURRENT_TIMESTAMP WHERE starttime = @starttime

  GOTO EndOfProc

ProcFailed:
  IF @log_mode = 1
    UPDATE PendingDeleteLog SET err = @err, rc = @rc, endtime = CURRENT_TIMESTAMP WHERE starttime = @starttime
  
EndOfProc:
  IF @log_mode = 1 -- If logging is on make sure we don't keep more than one day of data in PendingDeleteLog
    DELETE FROM PendingDeleteLog WHERE starttime < DATEADD(DAY, -1, CURRENT_TIMESTAMP)
  IF @verbose >= 1
  BEGIN
    SET @now = CONVERT(NVARCHAR, CURRENT_TIMESTAMP, 121)
    RAISERROR('%s: pa_processPendingDeletes: Finished. Return code = %d', 0, 1, @now, @rc) WITH NOWAIT
  END
  RETURN @rc
GO
